use std::fmt::Display;

#[derive(sqlx::Type)]
#[sqlx(transparent)]
#[repr(transparent)]
pub struct UserId(String);

impl UserId {
    pub fn new(name: &str, server_name: &str) -> anyhow::Result<Self> {
        let user_id = Self(format!("@{name}:{server_name}"));

        user_id.is_valid()?;

        Ok(user_id)
    }

    pub fn local_part(&self) -> &str {
        let col_idx = self.0.find(':').expect("Will always have at least one ':'");
        self.0[1..col_idx].as_ref()
    }

    fn is_valid(&self) -> anyhow::Result<()> {
        (self.0.len() <= 255)
            .then(|| ())
            .ok_or(UserIdError::TooLong(self.0.len()))?;

        let local_part = self.local_part();
        local_part
            .bytes()
            .all(|b| matches!(b, b'0'..=b'9' | b'a'..=b'z' | b'-' | b'.' | b'=' | b'_' | b'/'))
            .then(|| ())
            .ok_or(UserIdError::InvalidCharacters)?;

        Ok(())
    }
}

impl Display for UserId {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.0)
    }
}

#[derive(Debug, thiserror::Error)]
pub enum UserIdError {
    #[error("UserId too long {0} (expected < 255)")]
    TooLong(usize),
    #[error("Invalid character present in user id")]
    InvalidCharacters,
    #[error("Invalid UserId given")]
    Invalid,
}
